@@ -1,7 +1,5 @@
 import warnings
-from collections import defaultdict
 from itertools import combinations
-from itertools import permutations
 from itertools import product
 from typing import Dict
 from typing import List
@@ -84,7 +82,7 @@ class Conversation:
         text_column: str = "text",
         speaker_column: str = "speaker",
         id_column: Optional[str] = None,
-        language_model: str = "en_core_web_sm",
+        language_model: Union[str, spacy.Language] = "en_core_web_sm",
     ):
         # Set up data
         self.data = data.copy()
@@ -121,15 +119,25 @@ class Conversation:
         if isinstance(language_model, str):
             # We don't need full parsing by default, but do need sentence
             #   segmentation, which is faster
-            self.nlp = spacy.load(language_model)
+            self.nlp: spacy.Language = spacy.load(language_model)
             self.nlp.disable_pipe("parser")
             self.nlp.enable_pipe("senter")
-        else:
+        elif isinstance(language_model, spacy.Language):
             self.nlp = language_model
         self.data["spacy_doc"] = self._create_spacy_docs()
 
+    def __repr__(self):
+        if "name" in self.nlp.meta and "lang" in self.nlp.meta:
+            model_name = f"{self.nlp.meta['lang']}_{self.nlp.meta['name']}"
+        else:
+            model_name = "custom"
+        return (
+            f"Conversation({self.n_utterances} utterances, "
+            + f"{self.n_speakers} speakers, language_model='{model_name}')"
+        )
+
     def __str__(self):
-        return "Conversation object:\n" + str(self.data.head())
+        return repr(self)
 
     def _create_spacy_docs(self) -> pd.Series:
         """
@@ -144,6 +152,10 @@ class Conversation:
         # Regenerate spacy docs
         self.data["spacy_doc"] = self._create_spacy_docs()
 
+    @property
+    def n_utterances(self):
+        return self.data.shape[0]
+
     @property
     def n_speakers(self) -> int:
         """
@@ -306,8 +318,8 @@ class Conversation:
                 range_start = index + 1
                 range_end = range_start + recurrence_range
             elif direction == "backward":
-                range_start = index - recurrence_range
-                range_end = range_start + recurrence_range
+                range_start = max(index - recurrence_range, 0)
+                range_end = index
             similarity_vec = similarity.iloc[range_start:range_end, index]
             speaker_vec = speakers.iloc[range_start:range_end]
             if speaker == "self":
@@ -323,16 +335,26 @@ class Conversation:
         return recurrence
 
     def get_grouped_recurrence(
-        self, similarity: pd.DataFrame, grouping_column: str = "speaker"
+        self,
+        similarity: pd.DataFrame,
+        grouping_column: str = "speaker",
+        normalize: bool = True,
     ) -> pd.DataFrame:
         """Calculate overall recurrence between groups, e.g. person-to-person
         or group-to-group.
 
         See https://doi.org/10.1063/1.5024809
 
+        **Note:** the article suggests multiplying by the number of contributing
+        cells to normalize, but this seems odd, instead we divide by the
+        number of cells.
+
         Args:
             similarity: matrix of turn-turn similarity scores for the conversation,
                e.g. from
+            grouping_column: which column in data should be used to form the groups
+            normalize: Should recurrence score be normalized by the number
+              of contributing cells?
         """
         groups = self.data[grouping_column].unique()
         # Group for each text/turn in the conversation
@@ -352,16 +374,25 @@ class Conversation:
             np.triu(similarity, k=1), index=similarity.index, columns=similarity.columns
         )
 
-        pairs = permutations(groups, 2)
+        pairs = product(groups, repeat=2)
         for group_a, group_b in pairs:
             scores = similarity_upper.loc[
                 current_group == group_a, current_group == group_b
             ]
-            # Number of cells eligible to contribute is the total number with
-            #   group_i = a, group_j = b, and in the upper triangle where i < j
-            n_cells = in_upper_triangle.loc[scores.index, scores.columns].sum().sum()
-            recurrence.loc[group_a, group_b] = n_cells * scores.values.sum()
+            current_recurrence = scores.values.sum()
+            if normalize:
+                # Number of cells eligible to contribute is the total number with
+                #   group_i = a, group_j = b, and in the upper triangle where i < j
+                n_cells = (
+                    in_upper_triangle.loc[scores.index, scores.columns].sum().sum()
+                )
+                current_recurrence /= n_cells
+            recurrence.loc[group_a, group_b] = current_recurrence
 
+        recurrence = recurrence.fillna(0)
+        # Set names for index/columns to make scores easier to understand
+        recurrence.index.rename("initiated_by", inplace=True)
+        recurrence.columns.rename("repeated_by", inplace=True)
         return recurrence
 
 
@@ -503,6 +534,15 @@ class ConceptSimilarityModel(BaseSimilarityModel):
         super().__init__(conversation=conversation, key_terms=key_terms)
         self.sentence_window_size = sentence_window_size
 
+    def __repr__(self):
+        return (
+            f"ConceptSimilarityModel(key_terms={len(self.key_terms)}, "
+            + f"sentence_window_size={self.sentence_window_size})"
+        )
+
+    def __str__(self):
+        return repr(self)
+
     @staticmethod
     def _get_sentence_windows(doc: spacy.tokens.Doc, window_size: int):
         """
@@ -528,8 +568,22 @@ class ConceptSimilarityModel(BaseSimilarityModel):
             "cooccurrence": DataFrame of term-term cooccurrence counts,
             "total_windows":
         """
-        pair_counts = defaultdict(int)
-        occurrence_counts = defaultdict(int)
+        # vectorizer.terms_list is a property that's recomputed on access,
+        #   pull it out so we only access it once
+        all_terms = self.binary_vectorizer.terms_list
+        # Create series and dataframe to store occurrence counts
+        #   and cooccurrence counts
+        occurrence_series = pd.Series(
+            0,
+            index=all_terms,
+            dtype=pd.Int64Dtype(),
+        )
+        cooccurrence_matrix = pd.DataFrame(
+            0,
+            index=all_terms,
+            columns=all_terms,
+            dtype=pd.Int64Dtype(),
+        )
 
         total_windows = 0
         for doc in self.corpus:
@@ -543,28 +597,12 @@ class ConceptSimilarityModel(BaseSimilarityModel):
                     concat(self._filter_tokens(sentence) for sentence in window)
                 )
                 for token in current_tokens:
-                    occurrence_counts[token] += 1
+                    occurrence_series.at[token] += 1
 
-                pairs = combinations(sorted(current_tokens), 2)
-                for pair in pairs:
-                    pair_counts[pair] += 1
-
-        # Create series and dataframe to store occurrence counts
-        #   and cooccurrence counts
-        occurrence_series = pd.Series(
-            occurrence_counts,
-            index=self.binary_vectorizer.terms_list,
-            dtype=pd.Int64Dtype(),
-        )
-        cooccurrence_matrix = pd.DataFrame(
-            0,
-            index=self.binary_vectorizer.terms_list,
-            columns=self.binary_vectorizer.terms_list,
-            dtype=pd.Int64Dtype(),
-        )
-        for (term1, term2), count in pair_counts.items():
-            cooccurrence_matrix.loc[term1, term2] = count
-            cooccurrence_matrix.loc[term2, term1] = count
+                pairs = combinations(current_tokens, 2)
+                for term1, term2 in pairs:
+                    cooccurrence_matrix.at[term1, term2] += 1
+                    cooccurrence_matrix.at[term2, term1] += 1
 
         return {
             "occurrence": occurrence_series,
@@ -584,8 +622,10 @@ class ConceptSimilarityModel(BaseSimilarityModel):
             A dictionary mapping from 2-tuples ('i', 'j'), ('i', 'not_j'), etc. to a
             DataFrame of the counts of each.
         """
-        return {
+        raw_counts = {
             ("i", "j"): cooccurrence,
+            # NOTE: using some slightly tricky syntax to subtract the
+            #   cooccurrence from the occcurrence efficiently
             ("not_i", "not_j"): (
                 total_windows
                 - (
@@ -597,13 +637,10 @@ class ConceptSimilarityModel(BaseSimilarityModel):
             ("i", "not_j"): (-1 * cooccurrence).add(occurrence, axis="rows"),
             ("not_i", "j"): (-1 * cooccurrence).add(occurrence, axis="columns"),
         }
+        return raw_counts
 
-    @classmethod
     def get_term_similarity_matrix(
-        cls,
-        total_windows: int,
-        occurrence: pd.Series,
-        cooccurrence: pd.DataFrame,
+        self, cooccurrence_counts: Optional[dict] = None
     ) -> pd.DataFrame:
         """
         Calculate the similarity score S(t_i, t_j) between each pair of terms t_i.
@@ -611,15 +648,40 @@ class ConceptSimilarityModel(BaseSimilarityModel):
         See Angus (2012): https://doi.org/10/b49pvx
         for details of how to calculate these scores.
         """
-        contingency_counts = cls._get_contingency_counts(
-            total_windows=total_windows,
-            occurrence=occurrence,
-            cooccurrence=cooccurrence,
-        )
+        if cooccurrence_counts is None:
+            cooccurrence_counts = self.get_cooccurrence_counts()
+        contingency_counts = self._get_contingency_counts(**cooccurrence_counts)
+        occurrence = cooccurrence_counts["occurrence"]
+        total_windows = cooccurrence_counts["total_windows"]
         # Convert to probabilities
         contingency_probs: Dict[Tuple[str, str], pd.DataFrame] = {
             k: v / total_windows for k, v in contingency_counts.items()
         }
+        # Corrections
+        # See https://doi.org/10/b49pvx, p. 989 for details
+        # This count is used in multiple corrections so pull it out
+        ij_count = contingency_counts[("i", "j")]
+
+        # Correction for p(not_i, not_j)
+        count_plus_n = ij_count + total_windows
+        # Create a dataframe where the cell at (i, j) is O_i + O_j,
+        #   the total occurrence of i + j
+        total_occurrences = (
+            (ij_count * 0).add(occurrence, axis="rows").add(occurrence, axis="columns")
+        )
+        # NOTE: Having some issues with boolean indexing if the
+        #   boolean dataframe isn't explicitly cast to numpy bool
+        mask = (total_occurrences == count_plus_n).to_numpy(dtype="bool")
+        contingency_probs[("not_i", "not_j")][mask] = 1
+
+        # Correction for p(i, not_j)
+        # Occurrence of i equals cooccurrence ij
+        i_equals_ij = ij_count.eq(occurrence, axis="rows").to_numpy(dtype="bool")
+        contingency_probs[("i", "not_j")][i_equals_ij] = 1
+        # Correction for p(not_i, j)
+        # Occurrence of j equals cooccurrence ij
+        j_equals_ij = ij_count.eq(occurrence, axis="columns").to_numpy(dtype="bool")
+        contingency_probs[("not_i", "j")][j_equals_ij] = 1
 
         numerator = (
             contingency_probs[("i", "j")] * contingency_probs[("not_i", "not_j")]
@@ -629,22 +691,26 @@ class ConceptSimilarityModel(BaseSimilarityModel):
         )
 
         similarity_matrix = numerator / denominator
-        # NOTE: in order to divide invalid infinite values, we substitute
-        #  0.5 as the similarity value for values > 1. This occurs when
-        #  a term only occurs in a single context, such that the denominator
-        #  is zero
-        similarity_matrix = similarity_matrix.where(similarity_matrix <= 1, 0.5)
         return similarity_matrix
 
-    def get_concept_vectors(self, term_similarity_matrix: pd.DataFrame) -> pd.DataFrame:
+    def get_concept_vectors(
+        self, term_similarity_matrix: Optional[pd.DataFrame] = None
+    ) -> pd.DataFrame:
         """
         Get the feature vectors for each document.
 
+        Args:
+            term_similarity_matrix: Term-Term similarity matrix, as
+            returned by ConceptSimilarityModel.get_term_similarity_matrix.
+            If this is not provided, it will be recalculated, which
+            may be slow for larger datasets.
         Returns:
             A dataframe where the rows are the top terms in the corpus,
             and the columns are documents. Values are the feature value for
             that term's concept in each document.
         """
+        if term_similarity_matrix is None:
+            term_similarity_matrix = self.get_term_similarity_matrix()
         term_doc_df = pd.DataFrame(
             self.doc_term_matrix.T.todense(),
             index=self.binary_vectorizer.terms_list,
@@ -654,10 +720,7 @@ class ConceptSimilarityModel(BaseSimilarityModel):
         return key_term_similarity @ term_doc_df
 
     def get_conversation_similarity(self):
-        counts = self.get_cooccurrence_counts()
-        term_similarity_matrix = self.get_term_similarity_matrix(**counts)
-        concept_vectors = self.get_concept_vectors(term_similarity_matrix)
-
+        concept_vectors = self.get_concept_vectors()
         doc_doc_cosine = pd.DataFrame(
             pairwise.cosine_similarity(concept_vectors.T),
             index=self.conversation.data["text_id"],
